import gzip
import json
from time import time
from typing import Dict, Any

import boto3
import simplejson
from boto3.dynamodb.types import Binary

from sktmls import MLSClient, MLSENV, MLSRuntimeENV, MLSResponse, MLSClientError

MLS_DYNAMODB_API_URL = "/api/v1/dynamodb"

dynamodb = boto3.resource("dynamodb", region_name="ap-northeast-2")


class DynamoDBClient(MLSClient):
    """
    DynamoDB 관련 기능들을 제공하는 클라이언트입니다.
    """

    def __init__(
        self, env: MLSENV = None, runtime_env: MLSRuntimeENV = None, username: str = None, password: str = None
    ):
        """
        ## Args

        - env: (`sktmls.MLSENV`) 접근할 MLS 환경
        - runtime_env: (`sktmls.MLSRuntimeENV`) 클라이언트가 실행되는 환경 (`sktmls.MLSRuntimeENV.YE`|`sktmls.MLSRuntimeENV.EDD`|`sktmls.MLSRuntimeENV.LOCAL`) (기본값: `sktmls.MLSRuntimeENV.LOCAL`)
        - username: (str) MLS 계정명 (기본값: $MLS_USERNAME)
        - password: (str) MLS 계정 비밀번호 (기본값: $MLS_PASSWORD)

        ## Returns
        `sktmls.dynamodb.DynamoDBClient`

        ## Example

        ```python
        dynamodb_client = DynamoDBClient(env=MLSENV.STG, runtime_env=MLSRuntimeENV.YE, username="mls_account", password="mls_password")
        ```
        """
        super().__init__(env=env, runtime_env=runtime_env, username=username, password=password)

    def get_item(self, table_name: str, key: Dict[str, Any]) -> Dict[str, Any]:
        """
        주어진 Dynamodb 테이블의 특정 아이템을 조회합니다

        ## Args

        - table_name: (str) 테이블 명
        - key: (dict) 조회할 아이템의 key 값

        ## Returns

        (dict) 해당 아이템 값

        ## Example

        ```python
        item = dynamodb_client.get_item(table_name="sample_table", key={"item_id": "sample"})
        ```
        """
        if self._MLSClient__runtime_env == MLSRuntimeENV.MMS:
            table = dynamodb.Table(table_name)
            response = table.get_item(Key=key)
            item = response.get("Item")
            if not item:
                raise MLSClientError(code=404, msg="아이템이 존재하지 않습니다.")
            if (
                (table_name.startswith("user-profile") or table_name.startswith("profile-"))
                and not table_name.endswith("nogzip")
                and "values" in item
            ):
                values = json.loads(gzip.decompress(item["values"].value).decode())
                partition_key = "user_id" if table_name.startswith("user-profile") else "id"
                item = {partition_key: item[partition_key], **values}
            return {"item": json.loads(simplejson.dumps(item, use_decimal=True))}

        return self._request(
            method="POST", url=MLS_DYNAMODB_API_URL, data={"type": "get", "table_name": table_name, "body": key}
        ).results

    def put_item(self, table_name: str, item: Dict[str, Any]) -> Dict[str, Any]:
        """
        주어진 Dynamodb 테이블에 특정 아이템을 추가합니다

        ## Args

        - table_name: (str) 테이블 명
        - item: (dict) 추가할 아이템 값

        ## Returns

        (dict) 추가한 아이템 값

        ## Example
        ```python
        dynamodb_client.put_item(
            table_name="sample_table",
            item={"item_id": "sample", "sample_attribute": "sample_value"}
        )
        ```
        """
        if self._MLSClient__runtime_env == MLSRuntimeENV.MMS:
            table = dynamodb.Table(table_name)

            try:
                if (
                    table_name.startswith("user-profile") or table_name.startswith("profile-")
                ) and not table_name.endswith("nogzip"):
                    partition_key = "user_id" if table_name.startswith("user-profile") else "id"
                    compressed = gzip.compress(json.dumps(item).encode())
                    item = {partition_key: item[partition_key], "values": Binary(compressed)}

                if table_name.startswith("user-profile") or table_name.startswith("profile-"):
                    item["ttl"] = int(time() + 90 * 24 * 60 * 60)
                else:
                    item["ttl"] = int(time() + 30 * 24 * 60 * 60)

                response = table.put_item(Item=simplejson.loads(json.dumps(item), use_decimal=True))

                return response["ResponseMetadata"]
            except Exception:
                raise MLSClientError(code=404, msg="테이블이 존재하지 않습니다.")

        return self._request(
            method="POST", url=MLS_DYNAMODB_API_URL, data={"type": "put", "table_name": table_name, "body": item}
        ).results

    def update_item(self, table_name: str, item: Dict[str, Any]) -> Dict[str, Any]:
        """
        주어진 Dynamodb 테이블의 특정 아이템을 수정합니다

        ## Args

        - table_name: (str) 테이블 명
        - item: (dict) 수정할 아이템 값

        ## Returns

        (dict) 수정한 아이템 값

        ## Example

        ```python
        dynamodb_client.update_item(
            table_name="sample_table",
            item={"item_id": "sample", "sample_attribute": "sample_value_updated", "add_new_field": "added_field_value"}
        )
        ```
        """
        if self._MLSClient__runtime_env == MLSRuntimeENV.MMS:
            if "id" in item:
                partition_key = "id"
            elif "user_id" in item:
                partition_key = "user_id"
            else:
                partition_key = "item_id"
            updated_item = self.get_item(table_name, {partition_key: item[partition_key]})["item"]
            updated_item.update(item)
            return self.put_item(table_name, updated_item)

        return self._request(
            method="POST", url=MLS_DYNAMODB_API_URL, data={"type": "update", "table_name": table_name, "body": item}
        ).results

    def delete_item(self, table_name: str, key: Dict[str, Any]) -> MLSResponse:
        """
        주어진 Dynamodb 테이블의 특정 아이템을 삭제합니다

        ## Args

        - table_name: (str) 테이블 명
        - key: (dict) 삭제할 아이템의 key 값

        ## Returns

        `sktmls.MLSResponse`

        ## Example

        ```python
        dynamodb_client.delete_item(
            table_name="sample_table",
            item={"item_id": "sample"}
        )
        ```
        """
        return self._request(
            method="POST", url=MLS_DYNAMODB_API_URL, data={"type": "delete", "table_name": table_name, "body": key}
        )



from typing import Any, Dict, List

import numpy as np

from sktmls.models import MLSLightGBMModel


class SampleModel(MLSLightGBMModel):
    """
    단일 LightGBM을 사용하는 샘플 모델입니다.

    모델을 추론하여 y score가 0.2 이상인 경우 결과를 반환합니다.
    """

    def predict(self, x: List[Any], **kwargs) -> Dict[str, Any]:
        if not self.models:
            return {"items": []}

        try:
            y = self.models[0].predict(np.array([x])).flatten().tolist()[0]
        except Exception:
            return {"items": []}

        if y < 0.2:
            return {"items": []}

        return {
            "items": [{"id": "SAMPLE001", "name": "샘플상품", "props": {"context": "샘플컨텍스트", "score": y}, "type": "샘플타입"}]
        }



import math
import random
from calendar import monthrange
from datetime import datetime
from dateutil.relativedelta import relativedelta
from functools import reduce

import numpy as np
from haversine import haversine
from pytz import timezone

from sktmls import MLSClientError

TZ = timezone("Asia/Seoul")


class LogicProcessor:
    """
    상품 선택 로직 등을 프로세스하는 클래스입니다.

    아래 operations를 참조하여 json 형식의 로직을 구현하면 됩니다.

    ## 기본 문법

    기본적으로 JSON 형태의 로직은 아래 포맷을 따릅니다.

    ```python
    {
        "operator": [*args]
    }
    ```

    각각의 arg는 또 다른 JSON 형태의 로직이 될 수 있으며 recursive하게 프로세스됩니다.

    ```json
    {
        "and": [
            {"==": [3, 3]},
            {">": [
                5,
                {"max": [3, 10]}
            ]}
        ]
    }
    ```

    ## 피쳐 참조

    피쳐를 참조하는 방법은

    ```
    {
        "var": ["feature_name", 기본값]
    }
    ```

    형식입니다.

    리스트 또는 딕셔너리 타입의 피쳐 값을 참조하는 경우

    ```
    {
        "var": ["feature_name.0", 기본값]
    }
    ```

    ```
    {
        "var": ["feature_name.key_name", 기본값]
    }
    ```

    과 같은 형태로 하위 값에 접근이 가능합니다.

    ## 추가 키 참조

    Recommendation API v3 이후부터 지원하는 추가 키(세컨 키 등) 리스트는

    ```
    {
        "var": ["additional_keys.0", 기본값]
    }
    ```

    와 같이 `additional_key.index` 포맷으로 접근 가능합니다.

    ## 연산자

    - `==`: 두 값이 같은 경우 True. 타입을 고려하지 않음
    - `===`: 두 값이 같은 경우 True. 타입을 고려함
    - `!=`: 두 값이 다른 경우 True. 타입을 고려하지 않음
    - `!==`: 두 값이 다른 경우 True. 타입을 고려함
    - `<`: 왼쪽 값이 오른쪽 값보다 작은 경우 True
    - `<=`: 왼쪽 값이 오른쪽 값보다 작거나 같은 경우 True
    - `>`: 왼쪽 값이 오른쪽 값보다 큰 경우 True
    - `>=`: 왼쪽 값이 오른쪽 값보다 크거나 같은 경우 True
    - `!`: not
    - `and`: 모든 args가 True인 경우 True
    - `or`: 하나 이상의 args가 True인 경우 True
    - `any`: args[0]의 하나 이상의 element가 True인 경우 True
    - `all`: arg[0]의 모든 element가 True인 경우 True
    - `?:`: args[0]이 True이면 args[1] 반환, 아니면 args[2] 반환
    - `is`: args[0] is args[1]
    - `if`: args[0]이 True이면 args[1] 반환, 아닌 경우 만약 args[2]이 True이면 args[3] 반환, ... , 조건이 모두 False면 args[-1] 반환
    - `if_min`: args[0], args[2], args[4]... 중 args[i]가 최소인 경우 args[i + 1] 반환
    - `if_max`: args[0], args[2], args[4]... 중 args[i]가 최대인 경우 args[i + 1] 반환
    - `if_closest`: args[1], args[3], args[5]... 중 args[0]과 가장 가까운 값이 args[i]인 경우 args[i + 1] 반환
    - `if_closest_upper`: args[1], args[3], args[5]... 중 args[0]보다 크면서 가장 가까운 값이 args[i]인 경우 args[i + 1] 반환
    - `if_closest_lower`: args[1], args[3], args[5]... 중 args[0]보다 작으면서 가장 가까운 값이 args[i]인 경우 args[i + 1] 반환
    - `all_in`: args[0]의 element가 모두 args[1]에 포함된 경우 True
    - `any_in`: args[0]의 element가 하나라도 args[1]에 포함된 경우 True
    - `in`: 왼쪽 값이 오른쪽 값에 포함된 경우 True
    - `cat`: args들을 문자열 join
    - `join`: 리스트 args[0]을 args[1]을 구분자로 하여 문자열로 join
    - `split`: 문자열 args[0]을 args[1]을 구분자로 하여 분할
    - `+`: args들의 합
    - `*`: args들의 곱
    - `-`: 왼쪽 빼기 오른쪽. args의 길이가 1인 경우 -args[0]
    - `/`: 왼쪽 나누기 오른쪽
    - `%`: 왼쪽 나누기 오른쪽한 나머지
    - `inner`: arg[0]과 arg[1]의 내적 (둘 모두 1차원 배열인 경우 숫자를 반환하고 2차원 배열인 경우 리스트 반환)
    - `exp`: 지수함수 값 반환
    - `sigmoid`: Sigmoid 함수 값 반환
    - `min`: args 중 최소값
    - `max`: args 중 최대값
    - `argmin`: 최소값을 가지는 인덱스
    - `argmax`: 최대값을 가지는 인덱스
    - `get`: args[0]의 args[1]번째 element. 딕셔너리 타입의 경우 args[0]의 args[1] 키에 해당하는 값
    - `int`: int 파싱
    - `float`: float 파싱
    - `str`: str 파싱
    - `list`: args들을 리스트에 넣어 그대로 반환
    - `dict`: {args[0]: args[1], args[2]: args[3], ...}
    - `dict_keys`: 딕셔너리 타입의 args[0]의 키 리스트 반환
    - `dict_values`: 딕셔너리 타입의 args[0]의 값 리스트 반환
    - `pick`: args[0] 리스트 element 중 args[1]에서 정의한 인덱스만 뽑아냄
    - `pick_in`: args[0] 리스트 element 중 args[1]에 존재하는 것들만 필터
    - `remove`: args[1]에서 args[0] 값을 가진 element들을 삭제 후 반환
    - `remove_empty`: args[0]에서 빈 자료구조나 None인 element들을 삭제
    - `remove_duplicated`: 중복된 element들을 삭제. 순서가 앞인 element가 우선 (args[1]: 비교 index. None인 경우 element 자체를 비교).
    - `sample`: args[0] 중 args[1]개 만큼 랜덤 추출 (seed: args[2])
    - `first`: args[0] 중 첫 args[1]개를 반환
    - `replace`: args[2]의 element 중 args[0] 값을 args[1] 값으로 교체하여 반환
    - `repeat`: args[0]이 args[1]번 반복되는 리스트 반환
    - `len`: args[0]의 길이 반환
    - `empty`: args[0]의 길이가 0이면 True
    - `unique`: args[0] 리스트에서 중복값을 제거
    - `merge`: args들을 하나의 리스트로 합침
    - `vmerge`: 길이가 같은 리스트 args[0], args[1], args[2]... 의 같은 순서 element들끼리 묶어 반환
    - `count`: args의 길이
    - `abs`: args[0]의 절대값
    - `round`: 반올림 (args[0]을 args[1] 자리까지 반올림)
    - `mean`: 평균값 (리스트들이 args로 들어오는 경우 각 element의 평균값을 가지는 리스트 반환)
    - `closest`: args[1], args[2]... 중 args[0]과 가장 가까운 값
    - `closest_upper`: args[1], args[2]... 중 args[0]보다 크면서 가장 가까운 값
    - `closest_lower`: args[1], args[2]... 중 args[0]보다 작으면서 가장 가까운 값
    - `weekday`: 요일 인덱스 반환 (월요일 1 ~ 일요일 7)
    - `ndays`: 이번 달의 마지막 날짜
    - `day`: 오늘 날짜
    - `month`: 이번 달
    - `year`: 올해
    - `hour`: 시
    - `minute`: 분
    - `second`: 초
    - `microsecond`: 마이크로초
    - `now`: 현재 시각 반환 (datetime 객체)
    - `strptime`: 문자열을 시각으로 변환 (시각 문자열: args[0], 형식: args[1])
    - `strftime`: 시각을 문자열로 변환 (datetime 객체: args[0], 형식: args[1])
    - `timedelta`: 시간차 객체 반환 (relativedelta 객체) ([참조](https://dateutil.readthedocs.io/en/stable/relativedelta.html))
    - `dt+`: datetime 객체와 relativedelta 객체들의 합을 반환 (datetime 객체)
    - `km_to_latitude_degrees`: km 거리를 위도상 각도로 변환 (args[0]: km)
    - `km_to_longitude_degrees`: km 거리를 경도상 각도로 변환 (args[0]: km, args[1]: 기준 위도)
    - `distance`: 유클리디안 거리 계산 (args: x1, y1, x2, y2)
    - `geo_distance`: 위경도 기반 거리 계산 (args: lat1, lon1, lat2, lon2) (단위: km)
    - `rand`: 0~1 사이의 난수 (seed: args[0])
    - `beta`: 베타 분포를 따르는 random variable (a: args[0], b: args[1])
    - `pf`: 프로파일 API 조회 (api_type: args[0], profile_id: args[1], item_id/user_id: args[2], keys: args[3])
    - `with`: args[-1]의 프로세스에 {args[0]: arg[1], args[2]: args[3]...}의 데이터를 `var`로 참조 가능
    - `for`: args[0]를 iterator로 사용하여 args[1] 루프. {"var": "i"}, {"var": "i.0"} 등으로 iterator 참조 가능 (map과 유사)
    """

    def __init__(self):
        self.operations = {
            "==": self._soft_equals,
            "===": self._hard_equals,
            "!=": lambda a, b: not self._soft_equals(a, b),
            "!==": lambda a, b: not self._hard_equals(a, b),
            ">": lambda a, b: self._less(b, a),
            ">=": lambda a, b: self._less(b, a) or self._soft_equals(a, b),
            "<": self._less,
            "<=": self._less_or_equal,
            "!": lambda a: not a,
            "!!": bool,
            "and": lambda *args: reduce(lambda total, arg: total and arg, args, True),
            "or": lambda *args: reduce(lambda total, arg: total or arg, args, False),
            "all": lambda a: all(a),
            "any": lambda a: any(a),
            "?:": lambda a, b, c: b if a else c,
            "is": lambda a, b: a is b,
            "if": self._if,
            "if_min": self._if_min,
            "if_max": self._if_max,
            "if_closest": self._if_closest,
            "if_closest_upper": self._if_closest_upper,
            "if_closest_lower": self._if_closest_lower,
            "any_in": lambda a, b: any([i in b for i in a]),
            "all_in": lambda a, b: all([i in b for i in a]),
            "in": lambda a, b: a in b if "__contains__" in dir(b) else False,
            "contains": lambda a, b: b in a if "__contains__" in dir(a) else False,
            "cat": lambda *args: "".join(str(arg) for arg in args),
            "join": lambda a, sep: sep.join(str(i) for i in a),
            "split": lambda a, sep: [x.strip() for x in a.split(sep)],
            "+": self._plus,
            "*": lambda *args: reduce(lambda total, arg: total * float(arg), args, 1),
            "-": self._minus,
            "/": lambda a, b=None: a if b is None else float(a) / float(b),
            "%": lambda a, b: a % b,
            "inner": lambda a, b: np.inner(a, b).tolist(),
            "exp": lambda a: np.exp(a).tolist(),
            "sigmoid": lambda a: (1.0 / (1.0 + np.exp(np.negative(a)))).tolist(),
            "min": lambda *args: min(args),
            "max": lambda *args: max(args),
            "argmin": lambda *args: np.argmin(args),
            "argmax": lambda *args: np.argmax(args),
            "get": self._get,
            "int": self._int,
            "float": self._float,
            "str": self._str,
            "list": lambda *args: list(args),
            "dict": self._dict,
            "dict_keys": lambda a: list(a.keys()),
            "dict_values": lambda a: list(a.values()),
            "pick": lambda a, b: [a[i] for i in b],
            "pick_in": self._pick_in,
            "remove": lambda a, b: [i for i in b if i != a],
            "remove_empty": lambda a: [i for i in a if i],
            "remove_duplicated": self._remove_duplicated,
            "sample": self._sample,
            "first": lambda a, num_pick: list(a[:num_pick]),
            "replace": lambda a, b, c: [b if i == a else i for i in c],
            "repeat": lambda a, n: [a] * n,
            "len": lambda a: len(a),
            "empty": lambda a: len(a) == 0,
            "unique": self._unique,
            "merge": self._merge,
            "vmerge": self._vmerge,
            "count": lambda *args: sum(1 if a else 0 for a in args),
            "abs": lambda a: abs(a),
            "round": lambda a, b: round(a, b),
            "mean": lambda *args: np.mean(args, axis=0).tolist(),
            "closest": self._closest,
            "closest_upper": self._closest_upper,
            "closest_lower": self._closest_lower,
            "sort": lambda a, b: sorted(a, reverse=(b == "reverse")),
            "sort_by": lambda a, b, c: sorted(a, key=lambda x: x[b], reverse=(c == "reverse")),
            "weekday": lambda: datetime.now(TZ).isoweekday(),
            "ndays": lambda: monthrange(datetime.now(TZ).year, datetime.now(TZ).month)[1],
            "day": lambda: datetime.now(TZ).day,
            "month": lambda: datetime.now(TZ).month,
            "year": lambda: datetime.now(TZ).year,
            "hour": lambda: datetime.now(TZ).hour,
            "minute": lambda: datetime.now(TZ).minute,
            "second": lambda: datetime.now(TZ).second,
            "microsecond": lambda: datetime.now(TZ).microsecond,
            "now": lambda: datetime.now(TZ),
            "strptime": lambda a, b: TZ.localize(datetime.strptime(a, b)),
            "strftime": lambda a, b: datetime.strftime(a, b),
            "timedelta": lambda *args: relativedelta(**self._dict(*args)),
            "dt+": self._dt_plus,
            "km_to_latitude_degrees": lambda a: 360 * a / 40075,
            "km_to_longitude_degrees": lambda a, lat: 360 * a / 2 / math.pi / 6371 / math.cos(math.radians(lat)),
            "distance": lambda x1, y1, x2, y2: math.sqrt((x2 - x1) ** 2 + (y2 - y1) ** 2),
            "geo_distance": lambda lat1, long1, lat2, long2: haversine((lat1, long1), (lat2, long2)),
            "rand": self._rand,
            "beta": lambda a, b: np.random.beta(a, b),
            "return": lambda *args: args,
        }

    def _if_min(self, *args):
        min_arg = np.argmin([args[i] for i in range(0, len(args) - 1, 2)])
        return args[min_arg * 2 + 1]

    def _if_max(self, *args):
        max_arg = np.argmax([args[i] for i in range(0, len(args) - 1, 2)])
        return args[max_arg * 2 + 1]

    def _if_closest(self, arg, *comps):
        args = list(comps)
        for i in range(0, len(args) - 1, 2):
            args[i] = abs(arg - args[i])
        return self._if_min(*args)

    def _if_closest_upper(self, arg, *comps):
        args = []
        for i in range(0, len(comps) - 1, 2):
            if comps[i] > arg:
                args.append(comps[i])
                args.append(comps[i + 1])
        if not args:
            return None
        return self._if_closest(arg, *args)

    def _if_closest_lower(self, arg, *comps):
        args = []
        for i in range(0, len(comps) - 1, 2):
            if comps[i] < arg:
                args.append(comps[i])
                args.append(comps[i + 1])
        if not args:
            return None
        return self._if_closest(arg, *args)

    def _soft_equals(self, a, b):
        if isinstance(a, str) or isinstance(b, str):
            return str(a) == str(b)
        if isinstance(a, bool) or isinstance(b, bool):
            return bool(a) is bool(b)
        return a == b

    def _hard_equals(self, a, b):
        if type(a) != type(b):
            return False
        return a == b

    def _less(self, a, b, *args):
        types = set([type(a), type(b)])
        if float in types or int in types:
            try:
                a, b = float(a), float(b)
            except TypeError:
                return False
        return a < b and (not args or self._less(b, *args))

    def _less_or_equal(self, a, b, *args):
        return (self._less(a, b) or self._soft_equals(a, b)) and (not args or self._less_or_equal(b, *args))

    def _to_numeric(self, arg):
        if isinstance(arg, str):
            if "." in arg:
                return float(arg)
            else:
                return int(arg)
        return arg

    def _plus(self, *args):
        return sum(self._to_numeric(arg) for arg in args)

    def _minus(self, *args):
        if len(args) == 1:
            return -self._to_numeric(args[0])
        return self._to_numeric(args[0]) - self._to_numeric(args[1])

    def _closest(self, arg, *comps):
        closest = comps[0]
        min_diff = abs(arg - comps[0])
        for i in range(1, len(comps)):
            diff = abs(arg - comps[i])
            if min_diff > diff:
                min_diff = diff
                closest = comps[i]
        return closest

    def _closest_upper(self, arg, *comps):
        args = [c for c in comps if c > arg]
        if not args:
            return None
        return self._closest(arg, *args)

    def _closest_lower(self, arg, *comps):
        args = [c for c in comps if c < arg]
        if not args:
            return None
        return self._closest(arg, *args)

    def _get(self, a, b, not_found=None):
        try:
            return a[b]
        except (KeyError, TypeError, ValueError, IndexError):
            return not_found

    def _float(self, a):
        if isinstance(a, list) or isinstance(a, tuple):
            return [float(i) for i in a]
        return float(a)

    def _int(self, a):
        if isinstance(a, list) or isinstance(a, tuple):
            return [int(i) for i in a]
        return int(a)

    def _str(self, a):
        if isinstance(a, list) or isinstance(a, tuple):
            return [str(i) for i in a]
        return str(a)

    def _dict(self, *args):
        d = {}
        for i in range(0, len(args), 2):
            if i == len(args) - 1:
                break
            d[args[i]] = args[i + 1]
        return d

    def _pick_in(self, a, b):
        result = []
        for i in a:
            if i in b:
                result.append(i)
        return result

    def _remove_duplicated(self, a, b=None):
        result = []
        keys = []
        for i in a:
            comp = i[b] if b is not None else i
            if comp not in keys:
                keys.append(comp)
                result.append(i)
        return result

    def _dt_plus(self, dt, *args):
        if not isinstance(dt, datetime):
            raise ValueError("첫 arg가 datetime이 아닙니다.")
        for arg in args:
            if not isinstance(arg, relativedelta):
                raise ValueError("arg가 relativedelta가 아닙니다.")
            dt += arg
        return dt

    def _sample(self, a, num_pick, seed):
        random.seed(seed)
        return random.sample(a, num_pick)

    def _rand(self, seed):
        random.seed(seed)
        return random.random()

    def _unique(self, a):
        result = []
        for item in a:
            if item not in result:
                result.append(item)
        return result

    def _merge(self, *args):
        ret = []
        for arg in args:
            if isinstance(arg, list) or isinstance(arg, tuple):
                ret += list(arg)
            else:
                ret.append(arg)
        return ret

    def _vmerge(self, *args):
        ret = []
        for i in range(len(args[0])):
            r = []
            for j in range(len(args)):
                r.append(args[j][i])
            ret.append(r)
        return ret

    def _pf(self, pf_client, *args):
        if not pf_client:
            return None

        try:
            api_type = args[0]
            profile_id = args[1]
            target_id = args[2]
            keys = args[3]
        except (KeyError, TypeError, ValueError, IndexError):
            return None

        try:
            if api_type == "user":
                pf = pf_client.get_user_profile(profile_id=profile_id, user_id=target_id, keys=keys)
            elif api_type == "item":
                pf = pf_client.get_item_profile(profile_id=profile_id, item_id=target_id, keys=keys)
            else:
                pf = pf_client.get_profile(profile_id=profile_id, id=target_id, keys=keys)
        except MLSClientError:
            return None

        return [pf.get(key) for key in keys]

    def _graph(self, graph_client, *args):
        if not graph_client:
            return None

        try:
            request_type = args[0]
            params = self._dict(*args[1:])
        except (KeyError, TypeError, ValueError, IndexError):
            return None

        try:
            if request_type == "vertex":
                entities = graph_client.list_vertices(params)
            elif request_type == "edge":
                entities = graph_client.list_edges(params)
            else:
                return None
        except MLSClientError:
            return None

        return [e.get() for e in entities]

    def _dynamodb(self, dynamodb_client, *args):
        if not dynamodb_client:
            return None

        try:
            request_type = args[0]
            table_name = args[1]
            arg = self._dict(*args[2:])
        except (KeyError, TypeError, ValueError, IndexError):
            return None

        try:
            if request_type == "get":
                return dynamodb_client.get_item(table_name=table_name, key=arg)["item"]
            elif request_type == "put":
                return dynamodb_client.put_item(table_name=table_name, item=arg)
            elif request_type == "update":
                return dynamodb_client.update_item(table_name=table_name, item=arg)
            else:
                return None
        except MLSClientError:
            return None

    def _if(self, data, pf_client, graph_client, dynamodb_client, *args):
        for i in range(0, len(args) - 1, 2):
            if self.apply(args[i], data, pf_client, graph_client, dynamodb_client):
                return self.apply(args[i + 1], data, pf_client, graph_client, dynamodb_client)
        if len(args) % 2:
            return self.apply(args[-1], data, pf_client, graph_client, dynamodb_client)
        else:
            return None

    def _with(self, data, pf_client, graph_client, dynamodb_client, *args):
        if len(args) % 2 == 0:
            raise ValueError("with operator에 오류가 있습니다.")

        body = args[-1]
        for i in range(0, len(args), 2):
            if i == len(args) - 1:
                break
            data[args[i]] = self.apply(args[i + 1], data, pf_client, graph_client, dynamodb_client)

        results = self.apply(body, data, pf_client, graph_client, dynamodb_client)

        for i in range(0, len(args), 2):
            if i == len(args) - 1:
                break
            data.pop(args[i], None)

        return results

    def _for(self, data, pf_client, graph_client, dynamodb_client, *args):
        results = []
        it = self.apply(args[0], data, pf_client, graph_client, dynamodb_client)
        for i in it:
            data["i"] = i
            results.append(self.apply(args[1], data, pf_client, graph_client, dynamodb_client))
        data.pop("i", None)
        return results

    def _get_var(self, data, var_name, not_found=None):
        try:
            for key in str(var_name).split("."):
                try:
                    data = data[key]
                except TypeError:
                    data = data[int(key)]
        except (KeyError, TypeError, ValueError, IndexError):
            return not_found
        else:
            return data

    def _missing(self, data, *args):
        not_found = object()
        if args and isinstance(args[0], list):
            args = args[0]
        ret = []
        for arg in args:
            if self._get_var(data, arg, not_found) is not_found:
                ret.append(arg)
        return ret

    def _missing_some(self, data, min_required, args):
        if min_required < 1:
            return []
        found = 0
        not_found = object()
        ret = []
        for arg in args:
            if self._get_var(data, arg, not_found) is not_found:
                ret.append(arg)
            else:
                found += 1
                if found >= min_required:
                    return []
        return ret

    def apply(self, tests, data=None, pf_client=None, graph_client=None, dynamodb_client=None):
        """
        ## Args

        - tests: (dict) 테스트 할 json 형식의 로직
        - data: (optional) (dict) 로직에서 참조할 데이터 dict (기본값: None)
        - pf_client: (optional) (`sktmls.apis.profile_api.MLSProfileAPIClient`) 로직에서 사용할 프로파일 API 클라이언트 (기본값: None)
        - graph_client: (optional) (`sktmls.apis.graph_api.MLSGraphAPIClient`) 로직에서 사용할 그래프 API 클라이언트 (기본값: None)
        - dynamodb_client: (optional) (`sktmls.dynamodb.DynamoDBClient`) 로직에서 사용할 DynamoDB 클라이언트 (기본값: None)
        """
        if tests is None or not isinstance(tests, dict):
            return tests

        operator = list(tests.keys())[0]
        data = data or {}
        values = tests[operator]

        if operator == "if":
            return self._if(data, pf_client, graph_client, dynamodb_client, *values)

        if operator == "with":
            return self._with(data, pf_client, graph_client, dynamodb_client, *values)

        if operator == "for":
            return self._for(data, pf_client, graph_client, dynamodb_client, *values)

        if not isinstance(values, list) and not isinstance(values, tuple):
            values = [values]

        values = [self.apply(val, data, pf_client, graph_client, dynamodb_client) for val in values]

        if operator == "pf":
            return self._pf(pf_client, *values)
        if operator == "graph":
            return self._graph(graph_client, *values)
        if operator == "dynamodb":
            return self._dynamodb(dynamodb_client, *values)
        if operator == "var":
            return self._get_var(data, *values)
        if operator == "missing":
            return self._missing(data, *values)
        if operator == "missing_some":
            return self._missing_some(data, *values)

        if operator not in self.operations:
            raise ValueError(f"없는 operator입니다: {operator}")

        return self.operations[operator](*values)



from sktmls.models.contrib import MABRankingRuleModel


class TestMABRankingRuleModel:
    def test_000_mab_valid_without_tday(self, mocker):
        mocker.patch(
            "sktmls.apis.MLSProfileAPIClient.get_item_profile",
            return_value={
                "card_ranking": {
                    "mp_point_popular_seg": [10, 10],
                    "mp_point_new_seg": [20, 15],
                    "mp_point_context_seg": [10, 3],
                }
            },
        )
        mocker.patch("numpy.random.beta", side_effect=[0.3, 0.4, 0.1])

        mrrm = MABRankingRuleModel(
            model_name="test_model",
            model_version="test_version",
            feature="seg_id",
            mab_item_profile_info={"profile_id": "test_profile", "key": "card_ranking"},
            item_info={"id": "test_id", "name": "test_name", "type": "test_type", "props_key": "ranking"},
            num_pick=0,
            random_cycle="daily",
        )
        results = mrrm.predict(["test_seg"])

        assert len(results["items"]) == 1
        assert results["items"][0]["id"] == "test_id"
        assert results["items"][0]["name"] == "test_name"
        assert results["items"][0]["type"] == "test_type"
        assert results["items"][0]["props"] == {
            "ranking": ["mp_point_new_seg", "mp_point_popular_seg", "mp_point_context_seg"]
        }

    def test_001_mab_valid_num_pick(self, mocker):
        mocker.patch(
            "sktmls.apis.MLSProfileAPIClient.get_item_profile",
            return_value={
                "card_ranking": {
                    "mp_point_popular_seg": [10, 10],
                    "mp_point_new_seg": [20, 15],
                    "mp_point_context_seg": [10, 3],
                }
            },
        )
        mocker.patch("numpy.random.beta", side_effect=[0.1, 0.2, 0.3])

        mrrm = MABRankingRuleModel(
            model_name="test_model",
            model_version="test_version",
            feature="seg_id",
            mab_item_profile_info={"profile_id": "test_profile", "key": "card_ranking"},
            item_info={"id": "test_id", "name": "test_name", "type": "test_type", "props_key": "ranking"},
            num_pick=2,
            random_cycle="daily",
        )
        results = mrrm.predict(["test_seg"])

        assert len(results["items"]) == 1
        assert results["items"][0]["id"] == "test_id"
        assert results["items"][0]["name"] == "test_name"
        assert results["items"][0]["type"] == "test_type"
        assert results["items"][0]["props"] == {"ranking": ["mp_point_context_seg", "mp_point_new_seg"]}



from sktmls.models.contrib import MbrCustomBenefitRandomModel


class TestMbrSimilarBenefitModel:
    def test_000_mbr_valid_1a(self):
        mcbrm = MbrCustomBenefitRandomModel(
            model_name="test_model",
            model_version="test_version",
            features=[
                "user_id",
                "available_benefit_activity",
                "available_benefit_asianfood",
                "available_benefit_beauty",
            ],
            categories=["mbr_activity", "mbr_asianfood", "mbr_beauty"],
            contexts={"37": ["97", "27", "32", "99", "26", "133", "84", "37", "15", "38", "85", "48", "47", "12"]},
            item_type="test_type",
            random_cycle="hourly",
            num_pick=10,
        )

        results = mcbrm.predict(
            [
                "1234567890",
                ["103", "21", "a3", "a4"],
                None,
                ["48", "104", "20", "4"],
            ]
        )

        assert len(results["items"]) == 1

        item = results["items"][0]
        assert item["id"] == "48"
        assert item["type"] == "test_type"
        assert item["props"]["context_id"] == "37"

    def test_001_mbr_valid_none(self):
        mcbrm = MbrCustomBenefitRandomModel(
            model_name="test_model",
            model_version="test_version",
            features=[
                "user_id",
                "available_benefit_activity",
                "available_benefit_asianfood",
                "available_benefit_beauty",
            ],
            categories=["mbr_activity", "mbr_asianfood", "mbr_beauty"],
            contexts={"10": [], "20": ["10", "11"]},
            item_type="test_type",
            num_pick=10,
        )

        results = mcbrm.predict(
            [
                "1234567890",
                ["103", "21", "a3", "a4"],
                None,
                ["48", "104", "20", "4"],
            ]
        )

        assert len(results["items"]) == 0



from typing import Any, Dict, List

from sktmls.models import MLSCatBoostModel, MLSModelError


class CatBoostPointwiseRankingModel(MLSCatBoostModel):
    """
    MLS 모델 레지스트리에 등록되는 CatBoost 기반 클래스입니다.

    여러 차례 predict_proba를 수행하여 score가 높은 상품을 추천하는 모델입니다.
    """

    def __init__(
        self,
        model,
        model_name: str,
        model_version: str,
        point_feature: str,
        model_features: List[str],
        context_features: List[str],
        products: Dict[str, Any],
        item_type: str,
        cutoff: float,
        num_pick: int = 0,
    ):
        """
        ## Args

        - model: CatBoost로 학습한 모델 객체
        - model_name: (str) 모델 이름
        - model_version: (str) 모델 버전
        - point_feature: (str) 포인트로 사용되는 값 리스트를 가리키는 피쳐 이름
        - model_features: (list(str)) 모델 피쳐 리스트
        - context_features: (list(str)) 컨텍스트 전용 피쳐 리스트
        - products: (dict) 추천 상품 정보. 형식은 아래 Example 참조
        - item_type: (str) 추천 상품 타입 (`type`에 들어갈 값)
        - cutoff: (float) 추천 기준 스코어
        - num_pick: (optional) (int) 추천할 상위 스코어 상품 수. 0인 경우 cutoff를 넘는 모든 상품 추천 (기본값: 0)

        ## Example

        ```python
        products = {
            "TTT0000000": {
                "name": "T상품1",
                "context_features": ["context_call_overuse"],
                "context_default": "context_default",
            },
            "TTT0000001": {
                "name": "T상품2",
                "context_features": ["app_use_traffic_healthcare_yn", "context_discount_benefit"],
                "context_default": "context_default",
            },
            "TTT0000002": {
                "name": "T상품3",
                "context_features": ["context_school_contact"],
                "context_default": "context_age1020",
            },
        }

        my_model_v1 = PointwiseRankingModel(
            model_name="my_model",
            model_version="v1",
            point_feature="reco_items",
            model_features=["feature1", "feature2", "feature3"],
            context_features=["context_check_bill", "context_school_contact", "context_antivirusapp_usage"],
            products=products,
            item_type="app_sk",
            cutoff=0.0,
            num_pick=3,
        )

        # point_feature + model_features + context_features 순서로 리스트를 구성하여 predict 함수 호출
        result = my_model_v1.predict([["TW50000002", "TW50000005"], 0.3, 0.4, 0.5])
        ```
        """
        assert isinstance(point_feature, str) and len(point_feature) > 0, "`point_feature`가 유효하지 않습니다."
        assert isinstance(model_features, list) and len(model_features) > 0, "`model_features`가 유효하지 않습니다."
        assert isinstance(context_features, list), "`context_features`가 유효하지 않습니다."
        assert isinstance(num_pick, int), "`num_pick`이 유효하지 않습니다."

        super().__init__(model, model_name, model_version, [point_feature] + model_features + context_features)

        self.point_feature = point_feature
        self.model_features = model_features
        self.context_features = context_features
        self.products = products
        self.item_type = item_type
        self.cutoff = cutoff
        self.num_pick = num_pick

    def predict(self, x: List[Any], **kwargs) -> Dict[str, Any]:
        point_feature_values = x[0]
        model_feature_values = x[1 : len(self.model_features) + 1]
        context_feature_values = x[-len(self.context_features) :] if self.context_features else []

        if not point_feature_values:
            return {"items": []}

        if not isinstance(point_feature_values, list):
            raise MLSModelError("PointwiseRankingModel: `point_feature`의 값이 리스트 타입이 아닙니다.")

        context_feature_map = {name: context_feature_values[i] for i, name in enumerate(self.context_features)}

        preprocessed_x = [[item_id] + model_feature_values for item_id in point_feature_values]
        y = self.models[0].predict_proba(preprocessed_x)

        items = []
        for i, proba in enumerate(y):
            score = proba[1]
            item_id = point_feature_values[i]
            if score >= self.cutoff:
                product = self.products.get(item_id, {})
                product_context_features = [
                    f for f in product.get("context_features", []) if context_feature_map.get(f, "N") == "Y"
                ]
                context_id = (
                    product_context_features[0]
                    if product_context_features
                    else product.get("context_default", "context_default")
                )
                items.append(
                    {
                        "id": item_id,
                        "name": product.get("name", "#"),
                        "type": self.item_type,
                        "score": score,
                        "props": {"context_id": context_id},
                    }
                )
        num_pick = min(self.num_pick, len(items)) if self.num_pick > 0 else len(items)
        items = sorted(items, key=lambda x: x["score"], reverse=True)[:num_pick]
        items = [{k: v for k, v in d.items() if k != "score"} for d in items]

        return {"items": items}



import numpy as np
from catboost import CatBoostClassifier

from sktmls.models.contrib import CatBoostPointwiseRankingModel


class TestCatBoostPointwiseRankingModel:
    def test_000_pointwise(self, mocker):
        cpr = CatBoostPointwiseRankingModel(
            model=CatBoostClassifier(custom_loss=["Accuracy"], random_seed=42, logging_level="Silent"),
            model_name="test_model",
            model_version="test_version",
            point_feature="test_point",
            model_features=["feature1", "feature2", "feature3"],
            context_features=["context1", "context2", "context3"],
            products={
                "PROD001": {"name": "상품001", "context_features": ["context1"], "context_default": "context_default"},
                "PROD002": {
                    "name": "상품002",
                    "context_features": ["context2", "context3"],
                    "context_default": "context_default",
                },
                "PROD003": {"name": "상품003", "context_features": ["context3"], "context_default": "context_age1020"},
                "PROD004": {
                    "name": "상품004",
                    "context_features": ["context1", "context2", "context3"],
                    "context_default": "context_default",
                },
            },
            item_type="test_type",
            cutoff=0.5,
        )

        mocker.patch(
            "catboost.CatBoostClassifier.predict_proba", return_value=np.array([[0.9, 0.1], [0.7, 0.3], [0.3, 0.7]])
        )

        assert cpr.predict([["PROD001", "PROD002", "PROD003"], 4, 5, 6, "Y", "Y", "Y"]) == {
            "items": [{"id": "PROD003", "name": "상품003", "type": "test_type", "props": {"context_id": "context3"}}]
        }

    def test_001_pointwise_not_exists_in_products(self, mocker):
        cpr = CatBoostPointwiseRankingModel(
            model=CatBoostClassifier(custom_loss=["Accuracy"], random_seed=42, logging_level="Silent"),
            model_name="test_model",
            model_version="test_version",
            point_feature="test_point",
            model_features=["feature1", "feature2", "feature3"],
            context_features=["context1", "context2", "context3"],
            products={
                "PROD001": {"name": "상품001", "context_features": ["context1"], "context_default": "context_default"},
                "PROD002": {
                    "name": "상품002",
                    "context_features": ["context2", "context3"],
                    "context_default": "context_default",
                },
                "PROD004": {
                    "name": "상품004",
                    "context_features": ["context1", "context2", "context3"],
                    "context_default": "context_default",
                },
            },
            item_type="test_type",
            cutoff=0.5,
        )

        mocker.patch(
            "catboost.CatBoostClassifier.predict_proba", return_value=np.array([[0.9, 0.1], [0.7, 0.3], [0.3, 0.7]])
        )

        assert cpr.predict([["PROD001", "PROD002", "PROD003"], 4, 5, 6, "Y", "Y", "Y"]) == {
            "items": [{"id": "PROD003", "name": "#", "type": "test_type", "props": {"context_id": "context_default"}}]
        }

    def test_002_pointwise_all_n(self, mocker):
        cpr = CatBoostPointwiseRankingModel(
            model=CatBoostClassifier(custom_loss=["Accuracy"], random_seed=42, logging_level="Silent"),
            model_name="test_model",
            model_version="test_version",
            point_feature="test_point",
            model_features=["feature1", "feature2", "feature3"],
            context_features=["context1", "context2", "context3"],
            products={
                "PROD001": {"name": "상품001", "context_features": ["context1"], "context_default": "context_default"},
                "PROD002": {
                    "name": "상품002",
                    "context_features": ["context2", "context3"],
                    "context_default": "context_default",
                },
                "PROD003": {"name": "상품003", "context_features": ["context3"], "context_default": "context_age1020"},
                "PROD004": {
                    "name": "상품004",
                    "context_features": ["context1", "context2", "context3"],
                    "context_default": "context_default",
                },
            },
            item_type="test_type",
            cutoff=0.5,
        )

        mocker.patch(
            "catboost.CatBoostClassifier.predict_proba", return_value=np.array([[0.9, 0.1], [0.7, 0.3], [0.3, 0.7]])
        )

        assert cpr.predict([["PROD001", "PROD002", "PROD003"], 4, 5, 6, "N", "N", "N"]) == {
            "items": [
                {"id": "PROD003", "name": "상품003", "type": "test_type", "props": {"context_id": "context_age1020"}}
            ]
        }

    def test_003_pointwise_all_cut(self, mocker):
        cpr = CatBoostPointwiseRankingModel(
            model=CatBoostClassifier(custom_loss=["Accuracy"], random_seed=42, logging_level="Silent"),
            model_name="test_model",
            model_version="test_version",
            point_feature="test_point",
            model_features=["feature1", "feature2", "feature3"],
            context_features=["context1", "context2", "context3"],
            products={
                "PROD001": {"name": "상품001", "context_features": ["context1"], "context_default": "context_default"},
                "PROD002": {
                    "name": "상품002",
                    "context_features": ["context2", "context3"],
                    "context_default": "context_default",
                },
                "PROD003": {"name": "상품003", "context_features": ["context3"], "context_default": "context_age1020"},
                "PROD004": {
                    "name": "상품004",
                    "context_features": ["context1", "context2", "context3"],
                    "context_default": "context_default",
                },
            },
            item_type="test_type",
            cutoff=0.5,
        )

        mocker.patch(
            "catboost.CatBoostClassifier.predict_proba", return_value=np.array([[0.9, 0.1], [0.7, 0.3], [0.6, 0.4]])
        )

        assert cpr.predict([["PROD001", "PROD002", "PROD003"], 4, 5, 6, "Y", "Y", "Y"]) == {"items": []}

    def test_004_pointwise_all_in(self, mocker):
        cpr = CatBoostPointwiseRankingModel(
            model=CatBoostClassifier(custom_loss=["Accuracy"], random_seed=42, logging_level="Silent"),
            model_name="test_model",
            model_version="test_version",
            point_feature="test_point",
            model_features=["feature1", "feature2", "feature3"],
            context_features=["context1", "context2", "context3"],
            products={
                "PROD001": {"name": "상품001", "context_features": ["context1"], "context_default": "context_default"},
                "PROD002": {
                    "name": "상품002",
                    "context_features": ["context2", "context3"],
                    "context_default": "context_default",
                },
                "PROD003": {"name": "상품003", "context_features": ["context3"], "context_default": "context_age1020"},
                "PROD004": {
                    "name": "상품004",
                    "context_features": ["context1", "context2", "context3"],
                    "context_default": "context_default",
                },
            },
            item_type="test_type",
            cutoff=0.5,
        )

        mocker.patch(
            "catboost.CatBoostClassifier.predict_proba", return_value=np.array([[0.2, 0.8], [0.2, 0.8], [0.3, 0.7]])
        )

        assert cpr.predict([["PROD001", "PROD002", "PROD003"], 4, 5, 6, "Y", "Y", "Y"]) == {
            "items": [
                {"id": "PROD001", "name": "상품001", "type": "test_type", "props": {"context_id": "context1"}},
                {"id": "PROD002", "name": "상품002", "type": "test_type", "props": {"context_id": "context2"}},
                {"id": "PROD003", "name": "상품003", "type": "test_type", "props": {"context_id": "context3"}},
            ]
        }

    def test_005_pointwise_all_in_num_pick(self, mocker):
        cpr = CatBoostPointwiseRankingModel(
            model=CatBoostClassifier(custom_loss=["Accuracy"], random_seed=42, logging_level="Silent"),
            model_name="test_model",
            model_version="test_version",
            point_feature="test_point",
            model_features=["feature1", "feature2", "feature3"],
            context_features=["context1", "context2", "context3"],
            products={
                "PROD001": {"name": "상품001", "context_features": ["context1"], "context_default": "context_default"},
                "PROD002": {
                    "name": "상품002",
                    "context_features": ["context2", "context3"],
                    "context_default": "context_default",
                },
                "PROD003": {"name": "상품003", "context_features": ["context3"], "context_default": "context_age1020"},
                "PROD004": {
                    "name": "상품004",
                    "context_features": ["context1", "context2", "context3"],
                    "context_default": "context_default",
                },
            },
            item_type="test_type",
            cutoff=0.5,
            num_pick=2,
        )

        mocker.patch(
            "catboost.CatBoostClassifier.predict_proba", return_value=np.array([[0.2, 0.8], [0.2, 0.8], [0.3, 0.7]])
        )

        assert cpr.predict([["PROD001", "PROD002", "PROD003"], 4, 5, 6, "Y", "Y", "Y"]) == {
            "items": [
                {"id": "PROD001", "name": "상품001", "type": "test_type", "props": {"context_id": "context1"}},
                {"id": "PROD002", "name": "상품002", "type": "test_type", "props": {"context_id": "context2"}},
            ]
        }



from sktmls.models.contrib import RandomPickModel, PeriodicRandomPickModel


class TestRandomPickModel:
    def test_000_random_pick_1(self):
        rpm = RandomPickModel(
            model_name="test_model",
            model_version="test_version",
            candidates=[
                {"id": "test_id1", "name": "test_name1", "type": "test_type", "props": {}},
                {"id": "test_id2", "name": "test_name2", "type": "test_type", "props": {}},
            ],
            num_pick=1,
        )

        assert rpm.predict(None) in [
            {"items": [{"id": "test_id1", "name": "test_name1", "type": "test_type", "props": {}}]},
            {"items": [{"id": "test_id2", "name": "test_name2", "type": "test_type", "props": {}}]},
        ]

    def test_001_random_pick_2(self):
        rpm = RandomPickModel(
            model_name="test_model",
            model_version="test_version",
            candidates=[
                {"id": "test_id1", "name": "test_name1", "type": "test_type", "props": {}},
                {"id": "test_id2", "name": "test_name2", "type": "test_type", "props": {}},
            ],
            num_pick=2,
        )

        result = rpm.predict(None)
        assert result == {
            "items": [
                {"id": "test_id1", "name": "test_name1", "type": "test_type", "props": {}},
                {"id": "test_id2", "name": "test_name2", "type": "test_type", "props": {}},
            ]
        } or result == {
            "items": [
                {"id": "test_id2", "name": "test_name2", "type": "test_type", "props": {}},
                {"id": "test_id1", "name": "test_name1", "type": "test_type", "props": {}},
            ]
        }


class TestPeriodicRandomPickModel:
    def test_000_periodic_random_pick_2(self):
        rpm = PeriodicRandomPickModel(
            model_name="test_model",
            model_version="test_version",
            candidates=[
                {"id": "test_id1", "name": "test_name1", "type": "test_type", "props": {}},
                {"id": "test_id2", "name": "test_name2", "type": "test_type", "props": {}},
                {"id": "test_id3", "name": "test_name3", "type": "test_type", "props": {}},
                {"id": "test_id4", "name": "test_name4", "type": "test_type", "props": {}},
                {"id": "test_id5", "name": "test_name5", "type": "test_type", "props": {}},
            ],
            num_pick=2,
            random_cycle="daily",
        )

        result1 = rpm.predict("1234567890")
        result2 = rpm.predict("1234567890")
        assert result1 == result2



import pytest

from sktmls.models.contrib import InfoUnpaidRuleModel


class TestInfoUnpaidRuleModels:
    @pytest.fixture(scope="class")
    def param(self):
        return {"model": InfoUnpaidRuleModel("test_model", "test_version", ["filter_col"])}

    def test_000_info_unpaid_no_return(self, param):
        assert param.get("model").predict(["N"]) == {"items": []}

    def test_001_info_unpaid_return(self, param):
        assert param.get("model").predict(["Y"]) == {
            "items": [{"id": "INF0000001", "name": "미납 요금 및 납부 안내", "type": "info_unpaid", "props": {}}]
        }



import numpy as np
import lightgbm

from sktmls import MLSRuntimeENV
from sktmls.apis import MLSProfileAPIClient, MLSGraphAPIClient
from sktmls.dynamodb import DynamoDBClient
from sktmls.models.contrib import GenericLogicModel


class TestGenericLogicModel:
    def test_000_generic_context_success(self, mocker):
        mocker.patch("lightgbm.basic.Booster.predict", return_value=np.array([0.6]))
        mocker.patch("sktmls.apis.MLSProfileAPIClient.get_item_profile", return_value={"test_dimension": 300})
        mocker.patch(
            "sktmls.apis.MLSGraphAPIClient.list_vertices",
            return_value=[{"id": "test001", "label": "testlabel", "properties": {"hello": "world"}}],
        )
        mocker.patch("sktmls.dynamodb.DynamoDBClient.get_item", return_value={"id": "ITEM0001", "age": 3})

        glm = GenericLogicModel(
            model=lightgbm.basic.Booster(train_set=lightgbm.Dataset(np.array([[1, 2]]), np.array([1]))),
            model_name="test_model",
            model_version="test_version",
            features=["feature1", "feature2", "feature3", "emb_feature1", "context_feature1", "context_feature2"],
            preprocess_logic={
                "merge": [
                    {"var": ["feature1", 0.0]},
                    {"if": [{"==": [{"var": ["feature2", "N"]}, "Y"]}, 1, 0]},
                    {"float": [{"var": ["feature3", "0.0"]}]},
                    {"pick": [{"var": ["emb_feature1", [0.0] * 4]}, [1, 3]]},
                    {"first": [{"pf": ["item", "test_profile", {"var": "additional_keys.0"}, ["test_dimension"]]}, 1]},
                ]
            },
            postprocess_logic={
                "if": [
                    {">": [{"var": "y"}, 0.5]},
                    {
                        "list": [
                            {
                                "dict": [
                                    "id",
                                    "PROD001",
                                    "name",
                                    "제품1",
                                    "type",
                                    "my_type",
                                    "props",
                                    {
                                        "dict": [
                                            "test_score",
                                            {"var": "y"},
                                            "context_id",
                                            {
                                                "if": [
                                                    {"==": [{"var": ["context_feature1", "N"]}, "Y"]},
                                                    "context_feature1",
                                                    {"==": [{"var": ["context_feature2", "N"]}, "Y"]},
                                                    "context_feature2",
                                                    "context_default",
                                                ]
                                            },
                                        ]
                                    },
                                ]
                            }
                        ]
                    },
                    {">": [{"var": "y"}, 0.3]},
                    {
                        "list": [
                            {
                                "dict": [
                                    "id",
                                    "PROD002",
                                    "name",
                                    "제품2",
                                    "type",
                                    "my_type",
                                    "props",
                                    {
                                        "dict": [
                                            "test_score",
                                            {"var": "y"},
                                            "context_id",
                                            {
                                                "if": [
                                                    {"==": [{"var": ["context_feature2", "N"]}, "Y"]},
                                                    "context_feature2",
                                                    {"==": [{"var": ["context_feature1", "N"]}, "Y"]},
                                                    "context_feature1",
                                                    "context_default",
                                                ]
                                            },
                                        ]
                                    },
                                ]
                            }
                        ]
                    },
                    {"list": []},
                ]
            },
            predict_fn="predict",
        )

        assert glm._preprocess(
            [2.0, "Y", "4.0", [1, 2, 3, 4], "Y", "Y"],
            [1],
            MLSProfileAPIClient(client_id="a", apikey="a"),
            MLSGraphAPIClient(),
            DynamoDBClient(runtime_env=MLSRuntimeENV.MMS),
        ) == [
            2.0,
            1.0,
            4.0,
            2.0,
            4.0,
            300.0,
        ]

        assert glm.predict(
            x=[2.0, "Y", "4.0", [1, 2, 3, 4], "Y", "Y"],
            keys=[1],
            pf_client=MLSProfileAPIClient(client_id="a", apikey="a"),
            graph_client=MLSGraphAPIClient(),
            dynamodb_client=DynamoDBClient(runtime_env=MLSRuntimeENV.MMS),
        ) == {
            "items": [
                {
                    "id": "PROD001",
                    "name": "제품1",
                    "type": "my_type",
                    "props": {"context_id": "context_feature1", "test_score": 0.6},
                }
            ]
        }

    def test_001_generic_context_success(self, mocker):
        mocker.patch("lightgbm.basic.Booster.predict", return_value=np.array([0.5]))
        mocker.patch("sktmls.apis.MLSProfileAPIClient.get_item_profile", return_value={"test_dimension": 300})
        mocker.patch(
            "sktmls.apis.MLSGraphAPIClient.list_vertices",
            return_value=[{"id": "test001", "label": "testlabel", "properties": {"hello": "world"}}],
        )
        mocker.patch("sktmls.dynamodb.DynamoDBClient.get_item", return_value={"id": "ITEM0001", "age": 3})

        glm = GenericLogicModel(
            model=lightgbm.basic.Booster(train_set=lightgbm.Dataset(np.array([[1, 2]]), np.array([1]))),
            model_name="test_model",
            model_version="test_version",
            features=["feature1", "feature2", "feature3", "emb_feature1", "context_feature1", "context_feature2"],
            preprocess_logic={
                "merge": [
                    {"var": ["feature1", 0.0]},
                    {"if": [{"==": [{"var": ["feature2", "N"]}, "Y"]}, 1, 0]},
                    {"float": [{"var": ["feature3", "0.0"]}]},
                    {"pick": [{"var": ["emb_feature1", [0.0] * 4]}, [1, 3]]},
                    {"first": [{"pf": ["item", "test_profile", {"var": "additional_keys.0"}, ["test_dimension"]]}, 1]},
                ]
            },
            postprocess_logic={
                "if": [
                    {">": [{"var": "y"}, 0.5]},
                    {
                        "list": [
                            {
                                "dict": [
                                    "id",
                                    "PROD001",
                                    "name",
                                    "제품1",
                                    "type",
                                    "my_type",
                                    "props",
                                    {
                                        "dict": [
                                            "test_score",
                                            {"var": "y"},
                                            "context_id",
                                            {
                                                "if": [
                                                    {"==": [{"var": ["context_feature1", "N"]}, "Y"]},
                                                    "context_feature1",
                                                    {"==": [{"var": ["context_feature2", "N"]}, "Y"]},
                                                    "context_feature2",
                                                    "context_default",
                                                ]
                                            },
                                        ]
                                    },
                                ]
                            }
                        ]
                    },
                    {">": [{"var": "y"}, 0.3]},
                    {
                        "list": [
                            {
                                "dict": [
                                    "id",
                                    "PROD002",
                                    "name",
                                    "제품2",
                                    "type",
                                    "my_type",
                                    "props",
                                    {
                                        "dict": [
                                            "test_score",
                                            {"var": "y"},
                                            "context_id",
                                            {
                                                "if": [
                                                    {"==": [{"var": ["context_feature2", "N"]}, "Y"]},
                                                    "context_feature2",
                                                    {"==": [{"var": ["context_feature1", "N"]}, "Y"]},
                                                    "context_feature1",
                                                    "context_default",
                                                ]
                                            },
                                        ]
                                    },
                                ]
                            }
                        ]
                    },
                    {"list": []},
                ]
            },
            predict_fn="predict",
        )

        assert glm._preprocess(
            [2.0, "Y", "4.0", [1, 2, 3, 4], "Y", "Y"],
            [1],
            MLSProfileAPIClient(client_id="a", apikey="a"),
            MLSGraphAPIClient(),
            DynamoDBClient(runtime_env=MLSRuntimeENV.MMS),
        ) == [
            2.0,
            1.0,
            4.0,
            2.0,
            4.0,
            300.0,
        ]

        assert glm.predict(
            x=[2.0, "Y", "4.0", [1, 2, 3, 4], "Y", "Y"],
            keys=[1],
            pf_client=MLSProfileAPIClient(client_id="a", apikey="a"),
            graph_client=MLSGraphAPIClient(),
            dynamodb_client=DynamoDBClient(runtime_env=MLSRuntimeENV.MMS),
        ) == {
            "items": [
                {
                    "id": "PROD002",
                    "name": "제품2",
                    "type": "my_type",
                    "props": {"context_id": "context_feature2", "test_score": 0.5},
                }
            ]
        }

    def test_002_generic_context_success(self, mocker):
        mocker.patch("lightgbm.basic.Booster.predict", return_value=np.array([0.1]))
        mocker.patch("sktmls.apis.MLSProfileAPIClient.get_item_profile", return_value={"test_dimension": 300})
        mocker.patch(
            "sktmls.apis.MLSGraphAPIClient.list_vertices",
            return_value=[{"id": "test001", "label": "testlabel", "properties": {"hello": "world"}}],
        )
        mocker.patch("sktmls.dynamodb.DynamoDBClient.get_item", return_value={"id": "ITEM0001", "age": 3})

        glm = GenericLogicModel(
            model=lightgbm.basic.Booster(train_set=lightgbm.Dataset(np.array([[1, 2]]), np.array([1]))),
            model_name="test_model",
            model_version="test_version",
            features=["feature1", "feature2", "feature3", "emb_feature1", "context_feature1", "context_feature2"],
            preprocess_logic={
                "merge": [
                    {"var": ["feature1", 0.0]},
                    {"if": [{"==": [{"var": ["feature2", "N"]}, "Y"]}, 1, 0]},
                    {"float": [{"var": ["feature3", "0.0"]}]},
                    {"pick": [{"var": ["emb_feature1", [0.0] * 4]}, [1, 3]]},
                    {"first": [{"pf": ["item", "test_profile", {"var": "additional_keys.0"}, ["test_dimension"]]}, 1]},
                ]
            },
            postprocess_logic={
                "if": [
                    {">": [{"var": "y"}, 0.5]},
                    {
                        "list": [
                            {
                                "dict": [
                                    "id",
                                    "PROD001",
                                    "name",
                                    "제품1",
                                    "type",
                                    "my_type",
                                    "props",
                                    {
                                        "dict": [
                                            "test_score",
                                            {"var": "y"},
                                            "context_id",
                                            {
                                                "if": [
                                                    {"==": [{"var": ["context_feature1", "N"]}, "Y"]},
                                                    "context_feature1",
                                                    {"==": [{"var": ["context_feature2", "N"]}, "Y"]},
                                                    "context_feature2",
                                                    "context_default",
                                                ]
                                            },
                                        ]
                                    },
                                ]
                            }
                        ]
                    },
                    {">": [{"var": "y"}, 0.3]},
                    {
                        "list": [
                            {
                                "dict": [
                                    "id",
                                    "PROD002",
                                    "name",
                                    "제품2",
                                    "type",
                                    "my_type",
                                    "props",
                                    {
                                        "dict": [
                                            "test_score",
                                            {"var": "y"},
                                            "context_id",
                                            {
                                                "if": [
                                                    {"==": [{"var": ["context_feature2", "N"]}, "Y"]},
                                                    "context_feature2",
                                                    {"==": [{"var": ["context_feature1", "N"]}, "Y"]},
                                                    "context_feature1",
                                                    "context_default",
                                                ]
                                            },
                                        ]
                                    },
                                ]
                            }
                        ]
                    },
                    {"list": []},
                ]
            },
            predict_fn="predict",
        )

        assert glm._preprocess(
            [2.0, "Y", "4.0", [1, 2, 3, 4], "Y", "Y"],
            [1],
            MLSProfileAPIClient(client_id="a", apikey="a"),
            MLSGraphAPIClient(),
            DynamoDBClient(runtime_env=MLSRuntimeENV.MMS),
        ) == [
            2.0,
            1.0,
            4.0,
            2.0,
            4.0,
            300.0,
        ]

        assert (
            glm.predict(
                x=[2.0, "Y", "4.0", [1, 2, 3, 4], "Y", "Y"],
                keys=[1],
                pf_client=MLSProfileAPIClient(client_id="a", apikey="a"),
                graph_client=MLSGraphAPIClient(),
                dynamodb_client=DynamoDBClient(runtime_env=MLSRuntimeENV.MMS),
            )
            == {"items": []}
        )
